package org.apache.bcel.generic;

import org.apache.bcel.classfile.Utility;
import java.util.Collection;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

public class InstructionHandle
{
    public static final InstructionHandle[] EMPTY_ARRAY;
    static final InstructionTargeter[] EMPTY_INSTRUCTION_TARGETER_ARRAY;
    private InstructionHandle next;
    private InstructionHandle prev;
    private Instruction instruction;
    @Deprecated
    protected int i_position;
    private Set<InstructionTargeter> targeters;
    private Map<Object, Object> attributes;
    
    static InstructionHandle getInstructionHandle(final Instruction i) {
        return new InstructionHandle(i);
    }
    
    protected InstructionHandle(final Instruction i) {
        this.i_position = -1;
        this.setInstruction(i);
    }
    
    public void accept(final Visitor v) {
        this.instruction.accept(v);
    }
    
    public void addAttribute(final Object key, final Object attr) {
        if (this.attributes == null) {
            this.attributes = new HashMap<Object, Object>(3);
        }
        this.attributes.put(key, attr);
    }
    
    @Deprecated
    protected void addHandle() {
    }
    
    public void addTargeter(final InstructionTargeter t) {
        if (this.targeters == null) {
            this.targeters = new HashSet<InstructionTargeter>();
        }
        this.targeters.add(t);
    }
    
    void dispose() {
        final InstructionHandle instructionHandle = null;
        this.prev = instructionHandle;
        this.next = instructionHandle;
        this.instruction.dispose();
        this.instruction = null;
        this.i_position = -1;
        this.attributes = null;
        this.removeAllTargeters();
    }
    
    public Object getAttribute(final Object key) {
        if (this.attributes != null) {
            return this.attributes.get(key);
        }
        return null;
    }
    
    public Collection<Object> getAttributes() {
        if (this.attributes == null) {
            this.attributes = new HashMap<Object, Object>(3);
        }
        return this.attributes.values();
    }
    
    public final Instruction getInstruction() {
        return this.instruction;
    }
    
    public final InstructionHandle getNext() {
        return this.next;
    }
    
    public int getPosition() {
        return this.i_position;
    }
    
    public final InstructionHandle getPrev() {
        return this.prev;
    }
    
    public InstructionTargeter[] getTargeters() {
        if (!this.hasTargeters()) {
            return InstructionHandle.EMPTY_INSTRUCTION_TARGETER_ARRAY;
        }
        final InstructionTargeter[] t = new InstructionTargeter[this.targeters.size()];
        this.targeters.toArray(t);
        return t;
    }
    
    public boolean hasTargeters() {
        return this.targeters != null && !this.targeters.isEmpty();
    }
    
    public void removeAllTargeters() {
        if (this.targeters != null) {
            this.targeters.clear();
        }
    }
    
    public void removeAttribute(final Object key) {
        if (this.attributes != null) {
            this.attributes.remove(key);
        }
    }
    
    public void removeTargeter(final InstructionTargeter t) {
        if (this.targeters != null) {
            this.targeters.remove(t);
        }
    }
    
    public void setInstruction(final Instruction i) {
        if (i == null) {
            throw new ClassGenException("Assigning null to handle");
        }
        if (this.getClass() != BranchHandle.class && i instanceof BranchInstruction) {
            throw new ClassGenException("Assigning branch instruction " + i + " to plain handle");
        }
        if (this.instruction != null) {
            this.instruction.dispose();
        }
        this.instruction = i;
    }
    
    final InstructionHandle setNext(final InstructionHandle next) {
        return this.next = next;
    }
    
    void setPosition(final int pos) {
        this.i_position = pos;
    }
    
    final InstructionHandle setPrev(final InstructionHandle prev) {
        return this.prev = prev;
    }
    
    public Instruction swapInstruction(final Instruction i) {
        final Instruction oldInstruction = this.instruction;
        this.instruction = i;
        return oldInstruction;
    }
    
    @Override
    public String toString() {
        return this.toString(true);
    }
    
    public String toString(final boolean verbose) {
        return Utility.format(this.i_position, 4, false, ' ') + ": " + this.instruction.toString(verbose);
    }
    
    protected int updatePosition(final int offset, final int maxOffset) {
        this.i_position += offset;
        return 0;
    }
    
    static {
        EMPTY_ARRAY = new InstructionHandle[0];
        EMPTY_INSTRUCTION_TARGETER_ARRAY = new InstructionTargeter[0];
    }
}
